@@ -30,13 +30,14 @@ use crate::{
 use bls_signature_aggregator::Error as AggregatorError;
 use bytes::Bytes;
 use sn_messaging::{
-    client::Message as ClientMessage,
-    node::NodeMessage,
+    client::ClientMsg,
+    node::RoutingMsg,
     section_info::Error as TargetSectionError,
     section_info::{GetSectionResponse, Message as SectionInfoMsg, SectionInfo},
-    DstLocation, EndUser, MessageType,
+    DestInfo, DstLocation, EndUser, MessageType,
 };
 use std::{collections::BTreeSet, iter, net::SocketAddr};
+use xor_name::XorName;
 
 // Message handling
 impl Core {
@@ -44,6 +45,7 @@ impl Core {
         &mut self,
         sender: Option<SocketAddr>,
         msg: Message,
+        dest_info: DestInfo,
     ) -> Result<Vec<Command>> {
         let mut commands = vec![];
 
@@ -52,7 +54,9 @@ impl Core {
         if !in_dst_location || msg.dst().is_section() {
             // Relay closer to the destination or
             // broadcast to the rest of our section.
-            commands.extend(self.relay_message(&msg)?);
+            if let Some(cmds) = self.relay_message(&msg)? {
+                commands.push(cmds);
+            }
         }
         if !in_dst_location {
             // Message not for us.
@@ -68,12 +72,16 @@ impl Core {
         match self.decide_message_status(&msg)? {
             MessageStatus::Useful => {
                 trace!("Useful message from {:?}: {:?}", sender, msg);
-                commands.extend(self.update_section_knowledge(&msg)?);
-                commands.extend(self.handle_useful_message(sender, msg).await?);
+                let (entropy_commands, shall_be_handled) =
+                    self.check_for_entropy(&msg, dest_info.clone(), sender)?;
+                commands.extend(entropy_commands);
+                if shall_be_handled {
+                    commands.extend(self.handle_useful_message(sender, msg, dest_info).await?);
+                }
             }
             MessageStatus::Untrusted => {
                 debug!("Untrusted message from {:?}: {:?} ", sender, msg);
-                commands.push(self.handle_untrusted_message(sender, msg)?);
+                commands.push(self.handle_untrusted_message(sender, msg, dest_info)?);
             }
             MessageStatus::Useless => {
                 debug!("Useless message from {:?}: {:?}", sender, msg);
@@ -87,16 +95,23 @@ impl Core {
         &mut self,
         sender: SocketAddr,
         message: SectionInfoMsg,
+        dest_info: DestInfo, // The DestInfo contains the XorName of the sender and a random PK during the initial SectionQuery,
     ) -> Vec<Command> {
+        // Provide our PK as the dest PK, only redundant as the message itself contains details regarding relocation/registration.
+        let dest_info = DestInfo {
+            dest: dest_info.dest,
+            dest_section_pk: *self.section().chain().last_key(),
+        };
         match message {
-            SectionInfoMsg::GetSectionQuery(name) => {
+            SectionInfoMsg::GetSectionQuery(pk) => {
+                let name = XorName::from(pk);
                 debug!("Received GetSectionQuery({}) from {}", name, sender);
 
                 let response = if let (true, Ok(pk_set)) =
                     (self.section.prefix().matches(&name), self.public_key_set())
                 {
                     GetSectionResponse::Success(SectionInfo {
-                        prefix: self.section.authority_provider().prefix,
+                        prefix: self.section.authority_provider().prefix(),
                         pk_set,
                         elders: self
                             .section
@@ -109,21 +124,28 @@ impl Core {
                 } else {
                     // If we are elder, we should know a section that is closer to `name` that us.
                     // Otherwise redirect to our elders.
-                    let section = self
+                    let section_auth = self
                         .network
                         .closest(&name)
                         .unwrap_or_else(|| self.section.authority_provider());
-                    let addrs = section.elders.values().copied().collect();
-                    GetSectionResponse::Redirect(addrs)
+                    let targets = section_auth
+                        .elders()
+                        .iter()
+                        .map(|(name, addr)| (*name, *addr))
+                        .collect();
+                    GetSectionResponse::Redirect(targets)
                 };
 
                 let response = SectionInfoMsg::GetSectionResponse(response);
                 debug!("Sending {:?} to {}", response, sender);
 
                 vec![Command::SendMessage {
-                    recipients: vec![sender],
+                    recipients: vec![(name, sender)],
                     delivery_group_size: 1,
-                    message: MessageType::SectionInfo(response),
+                    message: MessageType::SectionInfo {
+                        msg: response,
+                        dest_info,
+                    },
                 }]
             }
             SectionInfoMsg::RegisterEndUserCmd {
@@ -145,16 +167,28 @@ impl Core {
                     ));
                 debug!("Sending {:?} to {}", response, sender);
 
+                let name = XorName::from(end_user);
                 vec![Command::SendMessage {
-                    recipients: vec![sender],
+                    recipients: vec![(name, sender)],
                     delivery_group_size: 1,
-                    message: MessageType::SectionInfo(response),
+                    message: MessageType::SectionInfo {
+                        msg: response,
+                        dest_info,
+                    },
                 }]
             }
             SectionInfoMsg::GetSectionResponse(_) => {
                 if let Some(RelocateState::InProgress(tx)) = &mut self.relocate_state {
                     trace!("Forwarding {:?} to the bootstrap task", message);
-                    let _ = tx.send((MessageType::SectionInfo(message), sender)).await;
+                    let _ = tx
+                        .send((
+                            MessageType::SectionInfo {
+                                msg: message,
+                                dest_info,
+                            },
+                            sender,
+                        ))
+                        .await;
                 }
                 vec![]
             }
@@ -172,7 +206,7 @@ impl Core {
     pub(crate) fn handle_timeout(&mut self, token: u64) -> Result<Vec<Command>> {
         self.dkg_voter
             .handle_timeout(&self.node.keypair, token)
-            .into_commands(&self.node)
+            .into_commands(&self.node, *self.section_chain().last_key())
     }
 
     // Insert the proposal into the proposal aggregator and handle it if aggregated.
@@ -215,7 +249,7 @@ impl Core {
 
     pub(crate) fn handle_dkg_failure(&mut self, proofs: DkgFailureProofSet) -> Result<Command> {
         let variant = Variant::DkgFailureAgreement(proofs);
-        let message = Message::single_src(&self.node, DstLocation::Direct, variant, None, None)?;
+        let message = Message::single_src(&self.node, DstLocation::Direct, variant, None)?;
         Ok(self.send_message_to_our_elders(message.to_bytes()))
     }
 
@@ -249,6 +283,7 @@ impl Core {
         &mut self,
         sender: Option<SocketAddr>,
         msg: Message,
+        dest_info: DestInfo,
     ) -> Result<Vec<Command>> {
         self.msg_filter.insert_incoming(&msg);
 
@@ -257,10 +292,21 @@ impl Core {
         } else {
             return Ok(vec![]);
         };
+        let src_name = msg.src().name();
 
         match msg.variant() {
-            Variant::OtherSection { section_auth, .. } => {
-                self.handle_other_section(section_auth.value.clone(), *msg.proof_chain_last_key()?)
+            Variant::SectionKnowledge { src_info, msg } => {
+                let src_info = src_info.clone();
+                self.update_section_knowledge(src_info.0, src_info.1);
+                if let Some(bounced_msg) = msg {
+                    Ok(vec![Command::HandleMessage {
+                        sender,
+                        message: *bounced_msg.clone(),
+                        dest_info,
+                    }])
+                } else {
+                    Ok(vec![])
+                }
             }
             Variant::Sync { section, network } => {
                 self.handle_sync(section.clone(), network.clone())
@@ -281,20 +327,36 @@ impl Core {
                 self.handle_join_request(msg.src().peer(sender)?, *join_request.clone())
             }
             Variant::UserMessage(content) => self.handle_user_message(&msg, content.clone()),
-            Variant::BouncedUntrustedMessage(message) => {
+            Variant::BouncedUntrustedMessage {
+                msg: bounced_msg,
+                dest_info,
+            } => {
                 let sender = sender.ok_or(Error::InvalidSrcLocation)?;
                 Ok(vec![self.handle_bounced_untrusted_message(
                     msg.src().peer(sender)?,
-                    msg.dst_key().copied(),
-                    *message.clone(),
+                    dest_info.dest_section_pk,
+                    *bounced_msg.clone(),
+                )?])
+            }
+            Variant::SectionKnowledgeQuery {
+                last_known_key,
+                msg: returned_msg,
+            } => {
+                let sender = sender.ok_or(Error::InvalidSrcLocation)?;
+                Ok(vec![self.handle_section_knowledge_query(
+                    *last_known_key,
+                    returned_msg.clone(),
+                    sender,
+                    src_name,
+                    msg.src().src_location().to_dst(),
                 )?])
             }
             Variant::DkgStart {
                 dkg_key,
-                section_auth,
-            } => self.handle_dkg_start(*dkg_key, section_auth.clone()),
+                elder_candidates,
+            } => self.handle_dkg_start(*dkg_key, elder_candidates.clone()),
             Variant::DkgMessage { dkg_key, message } => {
-                self.handle_dkg_message(*dkg_key, message.clone(), msg.src().name())
+                self.handle_dkg_message(*dkg_key, message.clone(), src_name)
             }
             Variant::DkgFailureObservation {
                 dkg_key,
@@ -312,7 +374,7 @@ impl Core {
                 let result = self.handle_proposal(content.clone(), proof_share.clone());
 
                 if let Some(addr) = sender {
-                    commands.extend(self.check_lagging(&addr, proof_share)?);
+                    commands.extend(self.check_lagging((src_name, addr), proof_share)?);
                 }
 
                 commands.extend(result?);
@@ -327,12 +389,21 @@ impl Core {
                 if let Some(RelocateState::InProgress(message_tx)) = &mut self.relocate_state {
                     if let Some(sender) = sender {
                         trace!("Forwarding {:?} to the bootstrap task", msg);
-                        let node_msg = NodeMessage::new(msg.to_bytes());
+                        let node_msg = RoutingMsg::new(msg.to_bytes());
                         let _ = message_tx
-                            .send((MessageType::NodeMessage(node_msg), sender))
+                            .send((
+                                MessageType::Routing {
+                                    msg: node_msg,
+                                    dest_info: DestInfo {
+                                        dest: src_name,
+                                        dest_section_pk: *self.section.chain().last_key(),
+                                    },
+                                },
+                                sender,
+                            ))
                             .await;
                     } else {
-                        error!("Missig sender of {:?}", msg);
+                        error!("Missing sender of {:?}", msg);
                     }
                 }
 
@@ -341,6 +412,39 @@ impl Core {
         }
     }
 
+    fn handle_section_knowledge_query(
+        &self,
+        given_key: Option<bls::PublicKey>,
+        msg: Box<Message>,
+        sender: SocketAddr,
+        src_name: XorName,
+        dst_location: DstLocation,
+    ) -> Result<Command> {
+        let chain = self.section.chain();
+        let given_key = if let Some(key) = given_key {
+            key
+        } else {
+            *self.section_chain().root_key()
+        };
+        let truncated_chain = chain.get_proof_chain_to_current(&given_key)?;
+        let section_auth = self.section.proven_authority_provider();
+        let variant = Variant::SectionKnowledge {
+            src_info: (section_auth.clone(), truncated_chain),
+            msg: Some(msg),
+        };
+
+        let msg = Message::single_src(self.node(), dst_location, variant, None)?;
+        let key = self.section_key_by_name(&src_name);
+        Ok(Command::send_message_to_node(
+            (src_name, sender),
+            msg.to_bytes(),
+            DestInfo {
+                dest: src_name,
+                dest_section_pk: *key,
+            },
+        ))
+    }
+
     pub(crate) fn verify_message(&self, msg: &Message) -> Result<bool> {
         let known_keys = self
             .section
@@ -359,46 +463,19 @@ impl Core {
         }
     }
 
-    pub(crate) fn handle_other_section(
-        &self,
-        section_auth: SectionAuthorityProvider,
-        src_key: bls::PublicKey,
-    ) -> Result<Vec<Command>> {
-        let mut commands = vec![];
-
-        if !self.network.has_key(&src_key) {
-            commands.extend(self.propose(Proposal::TheirKey {
-                prefix: section_auth.prefix,
-                key: src_key,
-            })?);
-        } else {
-            trace!(
-                "Ignore not new section key of {:?}: {:?}",
-                section_auth,
-                src_key
-            );
-            return Ok(commands);
-        }
-
-        commands.extend(self.propose(Proposal::SectionInfo(section_auth))?);
-
-        Ok(commands)
-    }
-
-    pub(crate) fn handle_user_message(
-        &mut self,
-        msg: &Message,
-        content: Bytes,
-    ) -> Result<Vec<Command>> {
+    fn handle_user_message(&mut self, msg: &Message, content: Bytes) -> Result<Vec<Command>> {
         trace!("handle user message {:?}", msg);
         if let DstLocation::EndUser(end_user) = msg.dst() {
+            let name = XorName::from(*end_user.id());
             let recipients = match end_user {
-                EndUser::AllClients(public_key) => {
-                    self.get_all_socket_addr(public_key).copied().collect()
-                }
+                EndUser::AllClients(public_key) => self
+                    .get_all_socket_addr(public_key)
+                    .copied()
+                    .map(|addr| (name, addr))
+                    .collect(),
                 EndUser::Client { socket_id, .. } => {
                     if let Some(socket_addr) = self.get_socket_addr(*socket_id).copied() {
-                        vec![socket_addr]
+                        vec![(name, socket_addr)]
                     } else {
                         vec![]
                     }
@@ -412,7 +489,13 @@ impl Core {
             return Ok(vec![Command::SendMessage {
                 recipients,
                 delivery_group_size: 1,
-                message: MessageType::ClientMessage(ClientMessage::from(content)?),
+                message: MessageType::Client {
+                    msg: ClientMsg::from(content)?,
+                    dest_info: DestInfo {
+                        dest: name,
+                        dest_section_pk: *self.section.chain().last_key(),
+                    },
+                },
             }]);
         }
 
@@ -474,6 +557,17 @@ impl Core {
         self.update_state(snapshot)
     }
 
+    pub fn handle_lagging_messages_on_sync(&mut self) -> Result<Vec<Command>> {
+        let mut commands = vec![];
+        let latest_key = *self.section_chain().last_key();
+        if let Some(lagged_commands) = self.lagging_messages.src_ahead.remove(&latest_key) {
+            // We now have the latest key, execute the messages that received when we were lagging.
+            commands.extend(lagged_commands);
+        }
+
+        Ok(commands)
+    }
+
     pub(crate) fn handle_join_request(
         &mut self,
         peer: Peer,
@@ -496,7 +590,11 @@ impl Core {
                 section_key: *self.section.chain().last_key(),
             };
             trace!("Sending {:?} to {}", variant, peer);
-            return Ok(vec![self.send_direct_message(peer.addr(), variant)?]);
+            return Ok(vec![self.send_direct_message(
+                (*peer.name(), *peer.addr()),
+                variant,
+                *self.section.chain().last_key(),
+            )?]);
         }
 
         if self.section.members().is_joined(peer.name()) {
